package cse.visit;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import polyglot.ast.Assign;
import polyglot.ast.Expr;
import polyglot.ast.IntLit;
import polyglot.ast.Term;
import polyglot.frontend.Job;
import polyglot.types.SemanticException;
import polyglot.visit.DataFlow;
import polyglot.visit.FlowGraph;
import polyglot.visit.FlowGraph.EdgeKey;
import polyglot.visit.FlowGraph.Peer;
import cse.extension.AssignExt;

/**
 * A dataflow analysis that collects all expressions of the form
 * * = 1; in the CFG. This is really an overkill just to illustrate how to use 
 * the DataFlow class for intraprocedural dataflow analyses.
 *  
 * @author akcheung
 *
 */
public class AssignOneDetector extends DataFlow<AssignOneDetector.FlowItem> 
{
	class FlowItem extends DataFlow.Item
	{
		Set<Assign> assignOneExprs;
		
		FlowItem () 
		{ 
			this.assignOneExprs = new HashSet<Assign>(); 
		}		
		
		FlowItem (Assign a)
		{
			this.assignOneExprs = new HashSet<Assign>();
			assignOneExprs.add(a);
		}
		
		FlowItem (Collection<FlowItem> others)
		{
			this.assignOneExprs = new HashSet<Assign>();
			for (FlowItem o : others)
				assignOneExprs.addAll(o.assignOneExprs);
		}
			
		/*
		 * Returns a new copy
		 */
		FlowItem add (Assign a)
		{
			FlowItem r = new FlowItem();
			r.assignOneExprs.addAll(this.assignOneExprs);
			r.assignOneExprs.add(a);
			return r;
		}
		
		@Override
		public boolean equals (Object o) 
		{
			return (o instanceof FlowItem) ? 
				this.assignOneExprs.equals( ((FlowItem)o).assignOneExprs) :
				false;			
		}

		@Override 
		public int hashCode ()
		{ 
			return assignOneExprs.hashCode(); 
		}
		
		@Override
		public String toString () 
		{
			StringBuilder sb = new StringBuilder();
			sb.append("exprs: ");
			for (Assign a : assignOneExprs)
				sb.append(a + " ");
			
			return sb.toString();
		}
	}
	
	public AssignOneDetector (Job job) 
	{
		super(job, job.extensionInfo().typeSystem(), job.extensionInfo().nodeFactory(), true);		
	}
	
	@Override
	protected FlowItem createInitialItem (FlowGraph<FlowItem> graph, Term node, boolean entry) 
	{
		return new FlowItem();
	}

	/** 
	 * Meet operation on the CFG -- union all flow items in this case
	 */
	@Override
	protected FlowItem confluence (List<FlowItem> inputFlowItems, Peer<FlowItem> meetNode,
								   FlowGraph<FlowItem> graph) 
	{
		return new FlowItem(inputFlowItems);
	}

	/**
	 * This is called on each node after dataflow is performed. After we are done
	 * with the entire graph (i.e., at each method's exit program point), we go through 
	 * all collected Assign expressions and set its assignToOne flag to true.
	 */
	@Override
	protected void check (FlowGraph<FlowItem> graph, Term node, boolean entry,
						  FlowItem inItem, Map<EdgeKey, FlowItem> outItems) throws SemanticException 
	{
		if (node == graph.root() && !entry)
		{
			// use the following if you want to dump the graph in DOT format
			// dumpFlowGraph(graph, node);
			
			for (Assign a : inItem.assignOneExprs)
				((AssignExt)a.ext()).assignToOne(true);
		}
	}
	
	/**
	 * This method performs dataflow at a given node and program point.
	 * Default is to just pass the same flow item to all successor nodes, unless if 
	 * we are processing an Assign expression with RHS == 1, then we create a new flow item and 
	 * insert the expression into it.
	 * 
	 * Call peer.isEntry() to check which program point the dataflow is performed. 
	 */
	@Override
    protected Map<EdgeKey, FlowItem> flow (FlowItem in, FlowGraph<FlowItem> graph, Peer<FlowItem> peer) 
    {
		FlowItem out = in;
		Term t = peer.node();
		if (t instanceof Assign)
		{
			Expr rhs = ((Assign)t).right();
			if (rhs instanceof IntLit && ((IntLit)rhs).value() == 1)
				out = in.add((Assign)t);
		}
		
		return itemToMap(out, peer.succEdgeKeys());
	}
}
